@@ -2,16 +2,13 @@ from typing import Type, Any, TypeVar, Callable, Generic, List, Optional, Dict,
 from abc import ABC, abstractmethod
 from dataclasses import dataclass, field
 from inspect import signature
+from copy import deepcopy
 
-from typing_inspect import get_origin, get_args
-
-
-class MappingException(Exception):
-    pass
+from typing_inspect import get_origin, get_args, is_union_type, is_forward_ref
 
 
 @dataclass
-class FieldMappingExceptionInfo:
+class MappingExceptionInfo:
     a: Type[Any]
     b: Type[Any]
     a_fields_chain: List[str] = field(default_factory=list)
@@ -21,25 +18,64 @@ class FieldMappingExceptionInfo:
         return 0 < len(self.a_fields_chain) or 0 < len(self.b_fields_chain)
 
 
+class MappingException(Exception):
+    def __init__(self, error_description: str, exc_info: Optional[MappingExceptionInfo] = None):
+        if exc_info is None:
+            message = error_description
+        else:
+            a_name = self._get_type_name(exc_info.a)
+            b_name = self._get_type_name(exc_info.b)
+            if exc_info.has_fields_chain():
+                a_field_name = self._get_filed_name(exc_info.a_fields_chain)
+                b_field_name = self._get_filed_name(exc_info.b_fields_chain)
+
+                message = (
+                    f"Cannot map field '{a_field_name}' of type '{a_name}' "
+                    f"to field '{b_field_name}' of type '{b_name}': {error_description}"
+                )
+            else:
+                message = f"Cannot map type '{a_name}' to type '{b_name}': {error_description}"
+
+        super(MappingException, self).__init__(message)
+
+    @staticmethod
+    def _get_type_name(t: Type) -> str:
+        if hasattr(t, "__name__"):
+            return t.__name__
+        else:
+            origin = get_origin(t)
+            if origin:
+                name = origin._name
+            else:
+                name = t._name
+            args = list(map(lambda x: MappingException._get_type_name(x), get_args(t)))
+            if len(args) != 0:
+                return f"{name}[{', '.join(args)}]"
+            else:
+                return name
+
+    @staticmethod
+    def _get_filed_name(fields_chain: List[str]):
+        return ".".join(fields_chain)
+
+
 class DuplicateMappingException(MappingException):
-    def __init__(self, left: Type, right: Type):
-        super(DuplicateMappingException, self).__init__(
-            f"Mapping from '{left.__name__}' to '{right.__name__}' already defined."
-        )
+    def __init__(self, exc_info: MappingExceptionInfo):
+        super(DuplicateMappingException, self).__init__("Mapping already defined.", exc_info)
 
 
 class MissingMappingException(MappingException):
-    def __init__(self, left: Type, right: Type):
+    def __init__(self, exc_info: MappingExceptionInfo, a: Type, b: Type):
+        a_name = MappingException._get_type_name(a)
+        b_name = MappingException._get_type_name(b)
         super(MissingMappingException, self).__init__(
-            f"Mapping from '{left.__name__}' to '{right.__name__}' is not defined."
+            f"Mapping from type '{a_name}' to type '{b_name}' is not defined.", exc_info
         )
 
 
 class ImproperlyConfiguredException(MappingException):
-    def __init__(self, left: Type, right: Type, error: str):
-        super(ImproperlyConfiguredException, self).__init__(
-            f"Mapping from '{left.__name__}' to '{right.__name__}' is improperly configured: {error}"
-        )
+    def __init__(self, exc_info: MappingExceptionInfo, error: str):
+        super(ImproperlyConfiguredException, self).__init__(f"Mapping is improperly configured: {error}.")
 
 
 class UnsupportedFieldException(MappingException):
@@ -48,17 +84,8 @@ class UnsupportedFieldException(MappingException):
 
 
 class FieldMappingException(MappingException):
-    def __init__(self, error: str, exc_info: Optional[FieldMappingExceptionInfo] = None):
-        if exc_info is None:
-            message = error
-        elif exc_info.has_fields_chain():
-            message = (
-                f"Cannot map field '{'.'.join(exc_info.a_fields_chain)}' of type '{exc_info.a.__name__}' to "
-                f"field '{'.'.join(exc_info.b_fields_chain)}' of type '{exc_info.b.__name__}': {error}"
-            )
-        else:
-            message = f"Cannot map type '{exc_info.a}' to type '{exc_info.b}': {error}"
-        super(MappingException, self).__init__(message)
+    def __init__(self, exc_info: MappingExceptionInfo, error: str):
+        super(MappingException, self).__init__(exc_info, error)
 
 
 T = TypeVar("T")
@@ -72,6 +99,7 @@ class FieldDescriptor(Generic[T, F]):
     getter: Callable[[T], F]
     setter: Optional[Callable[[T, F], None]]
     is_constructor_arg: bool
+    is_required_constructor_arg: bool
 
 
 T1 = TypeVar("T1")
@@ -85,7 +113,6 @@ class FieldMapRule(Generic[T1, F1, T2, F2]):
     from_field: FieldDescriptor[T1, F1]
     to_field: FieldDescriptor[T2, F2]
     converter: Optional[Callable[[F1], F2]]
-    check_types: bool
 
 
 class MappingDescriptor(ABC, Generic[T]):
@@ -105,6 +132,7 @@ class MappingDescriptor(ABC, Generic[T]):
                 getter=self.get_getter(field_name),
                 setter=self.get_setter(field_name),
                 is_constructor_arg=self.is_constructor_arg(field_name),
+                is_required_constructor_arg=self.is_required_constructor_arg(field_name),
             )
         else:
             return None
@@ -143,6 +171,12 @@ class MappingDescriptor(ABC, Generic[T]):
         """
         pass
 
+    def is_required_constructor_arg(self, field_name: str) -> bool:
+        """
+        Checks if field_name is required constructor arg
+        """
+        return field_name in self.get_required_constructor_args()
+
     @abstractmethod
     def get_declared_fields(self) -> Set[str]:
         """
@@ -287,7 +321,11 @@ class DictMappingDescriptor(MappingDescriptor):
         return True
 
 
-class MappingConfigFlow:
+L = TypeVar("L")
+R = TypeVar("R")
+
+
+class MappingConfigFlow(Generic[L, R]):
     def __init__(self, mapper: "Mapper", left_descriptor: MappingDescriptor, right_descriptor: MappingDescriptor):
         self.mapper = mapper
         self.left = left_descriptor.type
@@ -301,8 +339,11 @@ class MappingConfigFlow:
         self.r_to_l_touched = False
         self.r_to_l_map_list: List[FieldMapRule] = []
 
+        self.l_to_r_converter_callable: Optional[Callable[[L, Dict[str, Any]], R]] = None
+        self.r_to_l_converter_callable: Optional[Callable[[R, Dict[str, Any]], L]] = None
+
     def l_to_r(
-        self, left_field_name: str, right_field_name: str, converter: Callable[[Any], Any] = None, check_types=True
+        self, left_field_name: str, right_field_name: str, converter: Callable[[Any], Any] = None
     ) -> "MappingConfigFlow":
         left_field = self.left_descriptor.get_field_descriptor(left_field_name)
         if left_field is None:
@@ -312,14 +353,14 @@ class MappingConfigFlow:
         if right_field is None:
             raise UnsupportedFieldException(self.right_descriptor.type, right_field_name)
 
-        self.l_to_r_map_list.append(
-            FieldMapRule(from_field=left_field, to_field=right_field, converter=converter, check_types=check_types)
-        )
+        self.l_to_r_map_list.append(FieldMapRule(from_field=left_field, to_field=right_field, converter=converter))
+
         self.l_to_r_touched = True
+        self._l_to_r_check()
         return self
 
     def r_to_l(
-        self, left_field_name: str, right_field_name: str, converter: Callable[[Any], Any] = None, check_types=True
+        self, left_field_name: str, right_field_name: str, converter: Callable[[Any], Any] = None
     ) -> "MappingConfigFlow":
         left_field = self.left_descriptor.get_field_descriptor(left_field_name)
         if left_field is None:
@@ -329,27 +370,35 @@ class MappingConfigFlow:
         if right_field is None:
             raise UnsupportedFieldException(self.right_descriptor.type, right_field_name)
 
-        self.r_to_l_map_list.append(
-            FieldMapRule(from_field=right_field, to_field=left_field, converter=converter, check_types=check_types)
-        )
+        self.r_to_l_map_list.append(FieldMapRule(from_field=right_field, to_field=left_field, converter=converter))
+
         self.r_to_l_touched = True
+        self._r_to_l_check()
         return self
 
-    def bidirectional(self, l_field_name: str, r_field_name: str, check_types=True) -> "MappingConfigFlow":
-        self.l_to_r(l_field_name, r_field_name, check_types=check_types)
-        self.r_to_l(l_field_name, r_field_name, check_types=check_types)
+    def bidirectional(self, l_field_name: str, r_field_name: str) -> "MappingConfigFlow":
+        self.l_to_r(l_field_name, r_field_name)
+        self.r_to_l(l_field_name, r_field_name)
         return self
 
     def l_to_r_empty(self):
         if self.l_to_r_touched:
-            raise ImproperlyConfiguredException(self.left, self.right, "empty mapping after another configuration")
+            raise ImproperlyConfiguredException(
+                MappingExceptionInfo(self.left, self.right), "empty mapping after another configuration"
+            )
         self.l_to_r_touched = True
+
+        self._l_to_r_check()
         return self
 
     def r_to_l_empty(self):
         if self.r_to_l_touched:
-            raise ImproperlyConfiguredException(self.left, self.right, "empty mapping after another configuration")
+            raise ImproperlyConfiguredException(
+                MappingExceptionInfo(self.left, self.right), "empty mapping after another configuration"
+            )
+
         self.r_to_l_touched = True
+        self._r_to_l_check()
         return self
 
     def bidirectional_empty(self):
@@ -357,14 +406,15 @@ class MappingConfigFlow:
         self.l_to_r_empty()
         return self
 
-    def map_matching(self, ignore_case: bool = False, check_types=True) -> "MappingConfigFlow":
+    def map_matching(self, ignore_case: bool = False) -> "MappingConfigFlow":
         if self.left_descriptor.is_container_type() and self.right_descriptor.is_container_type():
             raise ImproperlyConfiguredException(
-                self.left, self.right, "map matching for two container types doesn't make sense"
+                MappingExceptionInfo(self.left, self.right), "map matching for two container types doesn't make sense"
             )
         if ignore_case and (self.left_descriptor.is_container_type() or self.right_descriptor.is_container_type()):
             raise ImproperlyConfiguredException(
-                self.left, self.right, "map matching for container types with ignored case does not supported yet"
+                MappingExceptionInfo(self.left, self.right),
+                "map matching for container types with ignored case does not supported yet",
             )
 
         if ignore_case:
@@ -385,7 +435,23 @@ class MappingConfigFlow:
         for f in common_fields:
             lf_name = l_fields[f]
             rf_name = r_fields[f]
-            self.bidirectional(lf_name, rf_name, check_types=check_types)
+            self.bidirectional(lf_name, rf_name)
+        return self
+
+    def l_to_r_converter(
+        self, converter: Union[Callable[[L], R], Callable[[L, Dict[str, Any]], R]]
+    ) -> "MappingConfigFlow":
+        self.l_to_r_converter_callable = self._wrap_converter_if_need_to(converter)
+
+        self._l_to_r_check()
+        return self
+
+    def r_to_l_converter(
+        self, converter: Union[Callable[[R], L], Callable[[R, Dict[str, Any]], L]]
+    ) -> "MappingConfigFlow":
+        self.r_to_l_converter_callable = self._wrap_converter_if_need_to(converter)
+
+        self._r_to_l_check()
         return self
 
     def register(self) -> None:
@@ -393,6 +459,33 @@ class MappingConfigFlow:
             self.mapper._add_map_rules(self.left, self.right, self.l_to_r_map_list)
         if self.r_to_l_touched:
             self.mapper._add_map_rules(self.right, self.left, self.r_to_l_map_list)
+        if self.l_to_r_converter_callable is not None:
+            self.mapper._add_converter(self.left, self.right, self.l_to_r_converter_callable)
+        if self.r_to_l_converter_callable is not None:
+            self.mapper._add_converter(self.right, self.left, self.r_to_l_converter_callable)
+
+    def _l_to_r_check(self):
+        if self.l_to_r_touched and self.l_to_r_converter_callable is not None:
+            raise ImproperlyConfiguredException(
+                MappingExceptionInfo(self.left, self.right), "Map rules and converter defined at the same time"
+            )
+
+    def _r_to_l_check(self):
+        if self.r_to_l_touched and self.r_to_l_converter_callable is not None:
+            raise ImproperlyConfiguredException(
+                MappingExceptionInfo(self.right, self.left), "Map rules and converter defined at the same time"
+            )
+
+    @staticmethod
+    def _wrap_converter_if_need_to(converter: Union[Callable[[T1], T2], Callable[[T1, Dict[str, Any]], T2]]):
+        if len(signature(converter).parameters) == 1:
+
+            def wrapped_converter(left: L, ignored_context: Dict[str, Any]):
+                return converter(left)
+
+            return wrapped_converter
+        else:
+            return converter
 
 
 class Mapper:
@@ -411,13 +504,17 @@ class Mapper:
     }
 
     def __init__(self, custom_descriptors: Optional[List[Type[MappingDescriptor]]] = None):
-        self.map_rules: Dict[Type, Dict[Type, List[FieldMapRule]]] = {}
         self.custom_descriptors = custom_descriptors if custom_descriptors else []
 
+        self.forward_ref_dict: Dict[str, Type[Any]] = {}
+
+        self.map_rules: Dict[Type, Dict[Type, List[FieldMapRule]]] = {}
+        self.converters: Dict[Type[Any], Dict[Type[Any], Callable[[Any, Dict[str, Any]], Any]]] = {}
+
     def mapping(self, a: Union[Type, MappingDescriptor], b: Union[Type, MappingDescriptor]) -> MappingConfigFlow:
-        if isinstance(a, Type):
+        if not isinstance(a, MappingDescriptor):
             a = self._wrap_type_to_descriptor(a)
-        if isinstance(b, Type):
+        if not isinstance(b, MappingDescriptor):
             b = self._wrap_type_to_descriptor(b)
         return MappingConfigFlow(self, a, b)
 
@@ -430,50 +527,110 @@ class Mapper:
 
     def _add_map_rules(self, a: Type, b: Type, rules: List[FieldMapRule]):
         a_type_mappings = self.map_rules.setdefault(a, {})
-        if b in a_type_mappings:
-            raise DuplicateMappingException(a, b)
+        a_type_converters = self.converters.setdefault(a, {})
+        if b in a_type_mappings or b in a_type_converters:
+            raise DuplicateMappingException(MappingExceptionInfo(a, b))
 
         a_type_mappings[b] = rules
+        self._add_class_to_forward_ref_dict(a)
+        self._add_class_to_forward_ref_dict(b)
 
-    def map(self, a_obj: Any, b: Type[T], exc_info: Optional[FieldMappingExceptionInfo] = None) -> T:
+    def _add_converter(self, a: Type[L], b: Type[R], converter: Callable[[L, Dict[str, Any]], R]):
+        a_type_mappings = self.map_rules.setdefault(a, {})
+        a_type_converters = self.converters.setdefault(a, {})
+        if b in a_type_mappings or b in a_type_converters:
+            raise DuplicateMappingException(MappingExceptionInfo(a, b))
+
+        a_type_converters[b] = converter
+        self._add_class_to_forward_ref_dict(a)
+        self._add_class_to_forward_ref_dict(b)
+
+    def _add_class_to_forward_ref_dict(self, t: Type):
+        if hasattr(t, "__name__"):
+            name = t.__name__
+            if name in self.forward_ref_dict and t != self.forward_ref_dict[name]:
+                raise Exception(
+                    f"Conflicting forward references '{name}'. Rearrange your class definitions or use type aliases."
+                )
+            else:
+                self.forward_ref_dict[name] = t
+
+    def _resolve_forward_ref(self, t: Type[Any]) -> Type[Any]:
+        if isinstance(t, str) or is_forward_ref(t):
+            if isinstance(t, str):
+                name = t
+            else:
+                name = get_args(t)[0]
+            if name in self.forward_ref_dict:
+                return self.forward_ref_dict[name]
+            else:
+                raise Exception(f"Unknown forward reference '{name}'")
+        else:
+            return t
+
+    def map(
+        self, a_obj: Any, b: Type[T], context: Dict[str, Any] = None, *, exc_info: Optional[MappingExceptionInfo] = None
+    ) -> T:
         a = a_obj.__class__
+        if context is None:
+            context = {}
         if exc_info is None:
-            exc_info = FieldMappingExceptionInfo(a, b)
+            exc_info = MappingExceptionInfo(a, b)
 
-        if a == b or b is Any:
-            return a_obj
-        elif self._has_mapping(a, b):
-            return self._map_with_map_rules(a_obj, b, exc_info)
+        if self._has_converter(a, b):
+            return self._convert_with_converter(a_obj, b, context, exc_info)
+        elif self._has_mapping_rules(a, b):
+            return self._map_with_map_rules(a_obj, b, context, exc_info)
         elif self._is_iterable_mapping_possible(a, b):
-            return self._map_iterables(a_obj, b, exc_info)
+            return self._map_iterables(a_obj, b, context, exc_info)
         elif self._has_primitive_mapping(a, b):
             return self._map_primitives(a_obj, b, exc_info)
+        elif self._is_direct_assignment_possible(a, b):
+            return deepcopy(a_obj)
         else:
-            raise MissingMappingException(a, b)
+            raise MissingMappingException(exc_info, a, b)
 
-    def _has_mapping(self, a: Type[Any], b: Type[Any]) -> bool:
+    def _has_converter(self, a: Type[Any], b: Type[Any]) -> bool:
+        return a in self.converters and b in self.converters[a]
+
+    def _convert_with_converter(
+        self, a_obj: Any, b: Type[Any], context: Dict[str, Any], exc_info: MappingExceptionInfo
+    ):
+        a = a_obj.__class__
+        try:
+            return self.converters[a][b](a_obj, context)
+        except Exception as e:
+            raise FieldMappingException(exc_info, "Error on converting") from e
+
+    def _has_mapping_rules(self, a: Type[Any], b: Type[Any]) -> bool:
         return a in self.map_rules and b in self.map_rules[a]
 
-    def _map_with_map_rules(self, a_obj: Any, b: Type[Any], exc_info: FieldMappingExceptionInfo):
+    def _map_with_map_rules(self, a_obj: Any, b: Type[Any], context: Dict[str, Any], exc_info: MappingExceptionInfo):
         a = a_obj.__class__
 
         constructor_args = {}
         fields = []
 
         for rule in self.map_rules[a][b]:
-            fields_exc_info = FieldMappingExceptionInfo(
-                rule.from_field.type,
-                rule.to_field.type,
+            from_field_type = self._resolve_forward_ref(rule.from_field.type)
+            to_field_type = self._resolve_forward_ref(rule.to_field.type)
+            fields_exc_info = MappingExceptionInfo(
+                from_field_type,
+                to_field_type,
                 exc_info.a_fields_chain + [rule.from_field.name],
                 exc_info.b_fields_chain + [rule.to_field.name],
             )
+            field_value = rule.from_field.getter(a_obj)
+            if field_value is None and not rule.to_field.is_required_constructor_arg:
+                continue
+
             if rule.converter is not None:
                 try:
-                    value = rule.converter(rule.from_field.getter(a_obj))
+                    value = rule.converter(field_value)
                 except Exception as e:
-                    raise FieldMappingException("Error on value conversion", fields_exc_info) from e
+                    raise FieldMappingException(fields_exc_info, "Error on value conversion") from e
             else:
-                value = self.map(rule.from_field.getter(a_obj), rule.to_field.type, fields_exc_info)
+                value = self.map(field_value, to_field_type, context, exc_info=fields_exc_info)
 
             if rule.to_field.is_constructor_arg:
                 constructor_args[rule.to_field.name] = value
@@ -490,18 +647,19 @@ class Mapper:
     def _has_primitive_mapping(self, a: Type[Any], b: Type[Any]) -> bool:
         return (a, b) in self.PRIMITIVE_CONVERTERS
 
-    def _map_primitives(self, a_obj: Any, b: Type[Any], exc_info: FieldMappingExceptionInfo):
+    def _map_primitives(self, a_obj: Any, b: Type[Any], exc_info: MappingExceptionInfo):
         a = a_obj.__class__
         primitive_converter = self.PRIMITIVE_CONVERTERS[(a, b)]
         try:
             return primitive_converter(a_obj)
         except Exception as e:
-            raise FieldMappingException("Exception on mapping primitive values", exc_info) from e
+            raise FieldMappingException(exc_info, "Exception on mapping primitive values") from e
 
     def _is_iterable_mapping_possible(self, a: Type[Any], b: Type[Any]) -> bool:
         return self._is_iterable(a) and self._is_iterable(b)
 
-    def _map_iterables(self, a_obj: Any, b: Type[Any], exc_info: FieldMappingExceptionInfo):
+    def _map_iterables(self, a_obj: Any, b: Type[Any], context: Dict[str, Any], exc_info: MappingExceptionInfo):
+        b = self._resolve_forward_ref(b)
         args = get_args(b)
         if len(args) == 0:
             # Iterable without type
@@ -510,19 +668,25 @@ class Mapper:
             try:
                 return to_type(*a_obj)
             except Exception as e:
-                raise FieldMappingException("Error on mapping iterable", exc_info) from e
+                raise FieldMappingException(exc_info, "Error on mapping iterable") from e
 
         elif len(args) == 1:
             # Iterable with type
             to_type = get_origin(b)
-            to_type_item = args[0]
+            to_type_item = self._resolve_forward_ref(args[0])
 
             mapped_list = []
             for index, item in enumerate(a_obj):
+                current_exc_info = MappingExceptionInfo(
+                    exc_info.a,
+                    exc_info.b,
+                    exc_info.a_fields_chain + [f"[{index}]"],
+                    exc_info.b_fields_chain + [f"[{index}]"],
+                )
                 try:
-                    mapped_list.append(self.map(item, to_type_item))
+                    mapped_list.append(self.map(item, to_type_item, context, exc_info=current_exc_info))
                 except Exception as e:
-                    raise FieldMappingException(f"Error on mapping iterable at index {index}", exc_info) from e
+                    raise FieldMappingException(exc_info, f"Error on mapping iterable at index {index}") from e
             return to_type(mapped_list)
 
         else:
@@ -531,15 +695,29 @@ class Mapper:
 
             mapped_list = []
             for index, item in enumerate(a_obj):
-                to_type_item = args[index]
-
+                to_type_item = self._resolve_forward_ref(args[index])
+                current_exc_info = MappingExceptionInfo(
+                    exc_info.a,
+                    exc_info.b,
+                    exc_info.a_fields_chain + [f"[{index}]"],
+                    exc_info.b_fields_chain + [f"[{index}]"],
+                )
                 try:
-                    mapped_list.append(self.map(item, to_type_item))
+                    mapped_list.append(self.map(item, to_type_item, context, exc_info=current_exc_info))
                 except Exception as e:
-                    raise FieldMappingException(f"Error on mapping iterable at index {index}", exc_info) from e
+                    raise FieldMappingException(exc_info, f"Error on mapping iterable at index {index}") from e
 
             return to_type(mapped_list)
 
+    def _is_direct_assignment_possible(self, a: Type[Any], b: Type[Any]) -> bool:
+        if b is Any:
+            return True
+        elif is_union_type(b):
+            return any([self._is_direct_assignment_possible(a, t) for t in get_args(b)])
+        elif issubclass(a, b):
+            return True
+        return False
+
     @staticmethod
     def _is_iterable(t: Type[Any]):
         origin = get_origin(t)
